@@ -31,8 +31,7 @@
 (require 'tramp)
 
 ;; Pacify byte-compiler.
-(eval-when-compile
-  (require 'custom))
+(declare-function ange-ftp-ftp-process-buffer "ange-ftp")
 (defvar ange-ftp-ftp-name-arg)
 (defvar ange-ftp-ftp-name-res)
 (defvar ange-ftp-name-format)
@@ -79,9 +78,9 @@ present for backward compatibility."
   ;;; This regexp recognizes absolute filenames with only one component
   ;;; on Windows, for the sake of hostname completion.
   (and (memq system-type '(ms-dos windows-nt))
-       (or (assoc "^[a-zA-Z]:/[^/:]*\\'" file-name-handler-alist)
+       (or (assoc "^[[:alpha:]]:/[^/:]*\\'" file-name-handler-alist)
 	   (setq file-name-handler-alist
-		 (cons '("^[a-zA-Z]:/[^/:]*\\'" .
+		 (cons '("^[:alpha:]]:/[^/:]*\\'" .
 			 ange-ftp-completion-hook-function)
 		       file-name-handler-alist)))))
 
@@ -127,7 +126,16 @@ pass to the OPERATION."
 	  ;; "ftp" method is used in the Tramp file name. So we unset
 	  ;; those values.
 	  (ange-ftp-ftp-name-arg "")
-	  (ange-ftp-ftp-name-res nil))
+	  (ange-ftp-ftp-name-res nil)
+	  (v (tramp-dissect-file-name
+	      (apply #'tramp-file-name-for-operation operation args) t)))
+      (setf (tramp-file-name-method v) tramp-ftp-method)
+      ;; Set "process-name" for thread support.
+      (tramp-set-connection-property
+       v "process-name"
+       (ange-ftp-ftp-process-buffer
+	(tramp-file-name-host v) (tramp-file-name-user v)))
+
       (cond
        ;; If argument is a symlink, `file-directory-p' and
        ;; `file-exists-p' call the traversed file recursively. So we
@@ -139,9 +147,7 @@ pass to the OPERATION."
        ;; "~/.netrc".
        ((memq operation '(file-directory-p file-exists-p))
 	(if (apply #'ange-ftp-hook-function operation args)
-	    (let ((v (tramp-dissect-file-name (car args) t)))
-	      (setf (tramp-file-name-method v) tramp-ftp-method)
-	      (tramp-set-connection-property v "started" t))
+	    (tramp-set-connection-property v "started" t)
 	  nil))
 
        ;; If the second argument of `copy-file' or `rename-file' is a
